<?php
class DromioWitness {
  function __construct($id,$text) {
    $this->id   = $id;
    $this->text = $text;
    $this->tokenAry = array();
    $this->taggedAry = array();
    $this->tokens = self::tokenize();
  }
  function getId() { return $this->id; }
  function getText() { return $this->text; }
  function getTokens() { return $this->tokens; }
  function getTokenAry() { return $this->tokenAry; }
  function getTaggedAry() { return $this->taggedAry; }
  function tokenize() {
    //for now, split by words; remove punctuation before matching
    //maybe later, split by punctuation too; but how to keep track of spaces?
    $tokens = array();
    $chars = str_split($this->text);
    $text = "";
    $tag = 0;
    foreach ($chars as $char) {
      if ($char == "<") { $tag++; }
      if ($char == ">") { $tag--; }
      if ($tag > 0 && $char == " ") {
        $text .= "_";
      } elseif ($tag > 0 && $char == "=") {
        $text .= ":";
      } elseif ($tag > 0 && $char == '"') {
        $text .= ":";
      } else {
        $text .= $char;
      }
    }

    //don't get confused when an ending tag prepends a new word
    while (preg_match('/[ ]*\\\\n[ ]*(\<\/.*?\>)/',$text)) {
      $text = preg_replace('/[ ]*\\\\n[ ]*(\<\/.*?\>)/'," $1\\n",$text);
    }
    while (preg_match('/ (\<\/.*?\>)/',$text)) {
      $text = preg_replace('/ (\<\/.*?\>)/',"$1 ",$text);
    }
    //or when an opening tag appends a word
    while (preg_replace('/(\<[^\/>]*?\>) /',$text)) {
      $text = preg_replace('/(\<[^\/>]*?\>) /'," $1",$text);
    }

    //if end tag is in the middle of a word ("<name>William Shakespeare</name>,")
    preg_match_all("/ ([^ ]+<\/[a-z]+>)([^\s\n\<]+)/",$text,$matchAry);
    if (count($matchAry) && count($matchAry[0])) {
      for ($i=0; $i<count($matchAry[0]); $i++) {
        $text = str_replace($matchAry[0][$i]," ".$matchAry[1][$i]." ".$matchAry[2][$i],$text);
      }
    }

    $words = preg_split("/\s/",$text);
    $i = 0;
    $count = count($words);
    $pad = 1 + floor(log10($count));
    if ($pad < 10) { $pad = 10; }
    $padding = "";
    for ($x=0;$x<$pad;$x++) { $padding .= "0"; }
    $tagsBefore = "";
    foreach ($words as $word) {
      if ($word) {
        $i++;
        $idx = substr("$padding$i",-1*$pad);

        $token = new stdClass();
        $token->id = $this->id."-".$idx;
        $token->taggedText = $word;
        $token->text = preg_replace('/\<.*?\>/','',$word);
        $token->tagsBefore = $tagsBefore;
        $token->tags = "";
        $token->taggedWord = $word;
        $token->tagsWord = "";
        $token->tagsChar = "";
        $token->tagsAfter = "";
        $token = self::parseTags($token,$tagsBefore);
        $tagsBefore = $token->tagsAfter;
        $token->taggedText = $word;
        $tokens[] = $token;
        $this->tokenAry[$token->id] = $token->text;
        $this->taggedAry[$token->id] = $token;
      }
    }
    return $tokens; 
  }
  function parseTags($token,$tagsBefore="") {
    $text = "";
    if (strlen($tagsBefore)) { 
      $text = "<m:$tagsBefore>";
      $text = str_replace(";","><m:",$text);
    }
    $text .= $token->taggedText;

    if (preg_match('/\<.*?\>/',$text,$match, PREG_OFFSET_CAPTURE)) {
      //for each tag, get full "tag" and the bare "element" name
      $tagText = $match[0][0];
      $tagpos = $match[0][1];
      $tag = substr($tagText,1,-1);
      $tagWord = "";
      $pregText = preg_quote($tagText,'/');
      $text = preg_replace("/$pregText/","",$text,1);
      if (strpos($tag,'_') > 0) { 
        $element = substr($tag,0,strpos($tag,'_'));
      } else {
        $element = $tag;
      }
      $tag = preg_replace("/^m:(m:)+/","m:",$tag);
      $element = preg_replace("/^(.:)*/","",$element);
      
      $close = strpos($text,"</$element>");
      //if its an opening tag, see if there's a closing one
      if (strpos($tag,'/') === 0) { //skip if is closing tag: </tag>
      } elseif (strpos($tag,'/') > 0) { //if is empty tag: <tag/>
        if (strlen($token->tags)) { $token->tags = "n:$tag;".$token->tags; }
        else { $token->tags = "n:$tag"; }
      } elseif (strpos($token->taggedWord,"<$tag>") > 0) { //not at beginning: wo<tag>rd</tag>
      } elseif ($close !== False && ($close+3+strlen($element) != strlen($text))) { //closing tag not at end: <tag>wo</tag>rd
        if (substr($tag,0,2) == "m:") {
          $token->taggedWord = preg_replace("/\<\/$element\>/","</f:$element>",$token->taggedWord,1);
        }
      } elseif ($close !== False) {
        $token->taggedWord = preg_replace("/\<$tag\>/","",$token->taggedWord,1);
        $token->taggedWord = preg_replace("/\<\/$element\>/","",$token->taggedWord,1);
        if ($tagpos == 0) {
          if (strlen($text) == strpos($text,"</$element>") + strlen("</$element>")) {
            $tagWord = $tagText;
          }
        }
        //closing tag found => f or y
        $text = preg_replace("/\<\/$element\>/","",$text,1);
        $newTag = "";
        if (substr($tag,0,2) == "m:") { $newTag = "f:".substr($tag,2); } else { $newTag .= "y:$tag"; }
        if (strlen($token->tags)) { $token->tags = "$newTag;".$token->tags; }
        else { $token->tags = $newTag; }
      } else {
        //no closing tag found => i or m
        $token->taggedWord = preg_replace("/\<$tag\>/","",$token->taggedWord,1);
        if (strlen($token->tagsAfter)) { $token->tagsAfter .= ";"; }
        $token->tagsAfter .= "$tag";
        if (substr($tag,0,2) == "m:") {
          $newTag = "$tag";
        } else {
          $newTag = "i:$tag";
        }
        if (strlen($token->tags)) { $newTag .= ";"; }
        $token->tags = "$newTag".$token->tags;
      }
    }
    $token->taggedText = $text;
    if (isset($tagWord)) {
      if (strlen($token->tagsWord)) { $token->tagsWord .= ";"; }
      $token->tagsWord .= $tagWord;
    }

    //continue to parse tags one by one until no more tags remain
    if (strpos($text,'<') !== False) { $token = self::parseTags($token); }
    return $token;
  }
  function echoText($text) {
    $text = str_replace("<","&lt;",$text);
    $text = str_replace(">","&gt;",$text);
    return $text;
  }
}
class DromioCollation {
  function __construct($textAry) {
    $this->iterationFlag = true;
    $this->collationFlag = "";
    $this->textAry = array();
    $this->matchAry = $textAry;
    $this->alignAry = array();
    $this->matchMatrix = array();
    $this->startTime = microtime(true);
    self::matchTexts($textAry);
    self::rankTexts($textAry);
    $this->alignAry = self::alignMatches($this->textAry);
  }
  function getMatches($i=0,$j=1) { return $this->matchMatrix[(string)$i][(string)$j]; }
  function getAlignment() { return $this->alignAry; }
  function getTexts() { return $this->textAry; }
  function rankTexts($textAry) {
    $rankAry = array();
    $translationAry = array();
    for ($i=0;$i<count($textAry);$i++) { 
      $score = 0;
      for ($j=0;$j<count($textAry);$j++) { 
        if ($i != $j) { $score += pow(count($this->matchMatrix[(string)$i][(string)$j]),2); }
      }
      $score = sqrt($score);
      $rankAry[$i] = $score;
    }
    arsort($rankAry);
    $newTextAry = array();
    $newMatchMatrix = array();
    $rank = 0;
    foreach ($rankAry as $key=>$val) { 
      $newTextAry[] = $textAry[$key];
      $newMatchMatrix[$key] = array();
      $translationAry[$key] = $rank;
      $rank++;
    }
    for ($i=0;$i<count($this->matchMatrix);$i++) { 
      for ($j=0;$j<count($this->matchMatrix);$j++) { 
        if ($i == $j) { continue; }
        $k = (string)$translationAry[$i];
        $l = (string)$translationAry[$j];
        $newMatchMatrix[$k][$l] = $this->matchMatrix[(string)$i][(string)$j];
      }
    }
    $this->textAry = $newTextAry;
    $this->matchMatrix = $newMatchMatrix;
  }
  function matchTexts($textAry) {
    if (count($textAry) == 0) {
    } elseif (count($textAry) == 1) {
      $this->matchAry = $textAry;
      $this->alignAry = $textAry;
    } else {
      for ($i=0;$i<count($textAry)-1;$i++) { 
        if (!isset($this->matchMatrix[(string)$i])) { $this->matchMatrix[(string)$i] = array(); }
        for ($j=$i+1;$j<count($textAry);$j++) { 
          if (!isset($this->matchMatrix[(string)$i][(string)$j])) { $this->matchMatrix[(string)$i][(string)$j] = array(); }
          $wit1 = $textAry[$i]->getTokens();
          $wit2 = $textAry[$j]->getTokens();
          $this->matchMatrix[(string)$i][(string)$j] = self::collate($wit1,$wit2);
          $this->matchMatrix[(string)$j][(string)$i] = array_flip($this->matchMatrix[(string)$i][(string)$j]);
        }
      }
    }
  }
  function collate($wit1,$wit2) {
    $matchAry = array();
    $this->collationFlag = "";
    $this->iterationFlag = true;

    $matchAry = self::iterateMatches($wit1,$wit2);
    //Case insensitive
    $this->collationFlag = "NC";
    $this->iterationFlag = true;
    $matchAry = self::iterateMatches($wit1,$wit2,$matchAry);
    //Case insensitive, no punctuation
    $this->collationFlag = "NC,NP";
    $this->iterationFlag = true;
    $matchAry = self::iterateMatches($wit1,$wit2,$matchAry);
    //Bring in non-unique words for small spans
    /*
    $this->collationFlag = "NU";
    $this->iterationFlag = true;
    $matchAry = self::iterateMatches($wit1,$wit2,$matchAry);
     */
    $this->collationFlag = "FINAL";
    $this->iterationFlag = true;
    $matchAry = self::iterateMatches($wit1,$wit2,$matchAry);
    return $matchAry;
  }
  function iterateMatches($witAry1,$witAry2,$matchAry = array()) {
    //if (microtime(true) - $this->startTime > 10) { return $matchAry; }
    $newMatches = array();
    $oldMisMatches = array();
    $idx = 0;
    $localStartTime = microtime(true);
    while ($this->iterationFlag) {
      if (microtime(true) - $localStartTime > 10) { return $matchAry; }
      $idx++;
      $matchAry = self::iterate($witAry1,$witAry2,$matchAry);
      $misMatches = self::checkMatches($matchAry);
      if (count($misMatches)) {
        if ($misMatches === $oldMisMatches) { $this->iterationFlag = false; }
        $oldMisMatches = $misMatches;
        foreach ($misMatches as $misMatch) { unset($matchAry[$misMatch]); }
      }
    }
    ksort($matchAry);
    return $matchAry;
  }
  function iterate($witAry1,$witAry2,$matchAry) {
    //loop through witAry1, witAry2, finding segments between matches
    //then try to match the segments
    $collationFlags = explode(',',$this->collationFlag);
    $segment1 = array();
    $segment2 = array();
    $witAry2 = array_reverse($witAry2);
    $wit2 = array_pop($witAry2);
    $wit2Matches = array_values($matchAry);
    $this->iterationFlag = false;
    $witCnt=0;
    foreach ($witAry1 as $wit1) {
      $witCnt++;
      $id1 = $wit1->id;
      $word1 = $wit1->text;
      if (isset($matchAry[$id1]) || $witCnt == count($witAry1)) {
        if (isset($matchAry[$id1])) { $match2 = $matchAry[$id1]; }
        else if (count($witAry2)) { $match2 = $witAry2[0]->id."z"; $segment1[] = $wit1; }
        else { $match2 = "ZZZZZ"; }
        while ($wit2 && $wit2->id < $match2 && (!in_array($wit2->id,$wit2Matches))) {
          if (!in_array($wit2->id,$wit2Matches)) { $segment2[] = $wit2; }
          $wit2 = array_pop($witAry2);
        }
        if (isset($wit2) && array_key_exists($id1,$matchAry) && $wit2->id == $matchAry[$id1]) { $wit2 = array_pop($witAry2); }

        if (in_array('OUT',$collationFlags)) { 
        } else {
          if (count($segment1) && count($segment2)) {
            if (in_array('FINAL',$collationFlags)) { 
              $newMatches = self::matchSegmentsFinal($segment1,$segment2);
            } else {
              $newMatches = self::matchSegments($segment1,$segment2);
            }
            if (count($newMatches)) { $this->iterationFlag = true; }
            $matchAry = array_merge($matchAry,$newMatches);
          }
        }
        $segment1 = array(); $segment2 = array();
      } else {
        $segment1[] = $wit1; 
      }
    }
    return $matchAry;
  }
  function matchSegments($segment1,$segment2) {
    $uniqueAry1 = self::getUniqueWords($segment1);
    $uniqueAry2 = self::getUniqueWords($segment2);
    $matchAry = self::matchUniqueWords($uniqueAry1,$uniqueAry2);
    $misMatches = self::checkMatches($matchAry);
    foreach ($misMatches as $misMatch) { unset($matchAry[$misMatch]); }
    return $matchAry;
  }
  function getUniqueWords($segment) {
    $words = array();
    $uniqueWords = array();
    foreach ($segment as $w) {
      $id = $w->id;
      $word = $w->text;

      $collationFlags = explode(',',$this->collationFlag);
      if ($word == "\\n") { $word = ""; }
      if (in_array('NC',$collationFlags)) { $word = strtoupper($word); }
      if (in_array('NP',$collationFlags)) { $newWord = preg_replace('/[^a-zA-Z0-9]/','',$word); if ($newWord) { $word = $newWord; }}

      if ($word == "") { }
      else if (isset($words[$word])) { $words[$word] .= ",".$id; }
      else { $words[$word] = $id; }
    }
    foreach ($words as $word => $ids) {
      if (strpos($ids,',') === false) { $uniqueWords[$word] = $ids; }
      elseif (count($segment) < 5 && in_array('NU',$collationFlags)) { 
        $ids = substr($ids,0,strpos($ids,','));
        $uniqueWords[$word] = $ids;
      }
    }
    return $uniqueWords;
  }
  function matchUniqueWords($Ary1,$Ary2) {
    $matchAry = array();
    foreach ($Ary1 as $word1 => $id1) {
      if (isset($Ary2[$word1])) {
        $matchAry[$id1] = $Ary2[$word1];
      }
    }
    return $matchAry;
  }
  function matchSegmentsFinal($segment1,$segment2) {
    $matchAry = array();
    if (count($segment1) == 1 && count($segment2) == 1) {
      $matchAry[$segment1[0]->id] = $segment2[0]->id;
    } else {
      $bestScore = 0; $besti = -1; $bestj = -1;
      $lookAheadMax = 3;
      for ($lookAhead1=0;$lookAhead1<$lookAheadMax;$lookAhead1++) { 
      for ($i=0;$i<count($segment1);$i++) { 
        $stri0 = $segment1[$i]->text;
        $stri = "";
        for ($idx=$i;$idx<count($segment1)&&$idx<=$i+$lookAhead1;$idx++) { $stri .= $segment1[$idx]->text; }
        for ($lookAhead2=0;$lookAhead2<$lookAheadMax;$lookAhead2++) { 
        for ($j=0;$j<count($segment2);$j++) { 
          $strj = "";
          for ($jdx=$j;$jdx<count($segment2)&&$jdx<=$j+$lookAhead2;$jdx++) { $strj .= $segment2[$jdx]->text; }
          if (strlen($stri) > 0 && strlen($strj) > 0 && strlen($stri) < 255 && strlen($strj) < 255) {
            $strLen = (strlen($stri) + strlen($strj)) / 2;
            $matchScore = 1 - (levenshtein($stri,$strj) / $strLen);
            if ($matchScore > $bestScore) { $bestScore = $matchScore; $besti = $i; $bestj = $j; }
            if ($matchScore == 2) {
              $i = count($segment1);
              $j = count($segment2);
              $lookAhead1 = $lookAheadMax;
              $lookAhead2 = $lookAheadMax;
            }
          }
        }}
      }}
      if ($besti > -1 && $bestj > -1) {
        $matchAry[$segment1[$besti]->id] = $segment2[$bestj]->id;
      }

    }
    return $matchAry;
  }
  function checkMatches($matchAry) {
    $rtnAry = array();
    $holdId = "";
    ksort($matchAry);
    foreach ($matchAry as $id1 => $id2) {
      if ($id2 < $holdId) {
        $rtnAry[] = array_search($holdId,$matchAry);
        $rtnAry[] = array_search($id2,$matchAry);
      }
      $holdId = $id2;
    }
    return $rtnAry;
  }
  function outputMatches($Ary) { }
  function alignMatches($textAry) {
    //Match 2 with 1; 3 with 1+2; 4 with 1+2+3
    $alignAry = array();
    foreach ($textAry[0]->getTokens() as $wit1) { $alignAry[] = $wit1->id; }
    for ($x=1;$x<count($textAry);$x++) { 
      $newAry = array();
      //reverse because pop faster than shift
      $witsX = array_reverse($textAry[$x]->getTokens());
      $witX = array_pop($witsX);
      foreach ($alignAry as $aligned) {
        $matchMin = "z";
        $matches = explode(",",$aligned);
        foreach ($matches as $match) {
          for ($y=0;$y<$x;$y++) { 
            if (isset($this->matchMatrix[(string)$y][(string)$x][$match])) {
              $matchX = $this->matchMatrix[(string)$y][(string)$x][$match];
              if ($matchX < $matchMin) { $matchMin = $matchX; }
            }
          }
        }
        if ($matchMin == 'z') { 
          $newAry[] = $aligned;
        } elseif (is_null($witX)) { 
          $newAry[] = $aligned;
        } elseif ($witX->id > $matchMin) { 
          $newAry[] = $aligned;
        } elseif ($witX->id == $matchMin) { 
          $newAry[] = "$aligned,$matchMin";
          $witX = array_pop($witsX);
        } elseif ($witX->id < $matchMin) { 
          while ($witX && $witX->id < $matchMin) {
            $newAry[] = $witX->id;
            $witX = array_pop($witsX);
          }
          $newAry[] = "$aligned,$matchMin";
          $witX = array_pop($witsX);
        } else { 
        }
      }
      if ($witX) { $newAry[] = $witX->id; }
      while (count($witsX)) {
        $witX = array_pop($witsX);
        $newAry[] = $witX->id;
      }
      $alignAry = $newAry;
    }
    return $alignAry;
  }
  function showTable($tagged=False) {
    $outText = "";
    $witAry = $this->getTexts();
    $matchAry = $this->getAlignment();

    $outText .= "<table border=1>\n";
    $outText .= "<tr>\n";
    for ($w=0;$w<count($witAry);$w++) {
      $wit = $witAry[$w];
      $witName = $wit->getId();
      $outText .= "<th>$witName</th>\n";
    }
    $outText .= "</tr>\n";
    for ($m=0;$m<count($matchAry);$m++) {
      $outText .= "<tr>\n";
      for ($w=0;$w<count($witAry);$w++) {
        $outText .= "<td>";
        $matches = $matchAry[$m];
        $keys = explode(",",$matches);
        foreach ($keys as $key) {
          $wit = $witAry[$w]->getTokenAry();
          $witTagged = $witAry[$w]->getTaggedAry();
          if (isset($wit[$key])) {
            $tag="";
            if (isset($witTagged[$key])) {
              if (strlen($witTagged[$key]->tags)) {
                $tag .= "<br/>".$witTagged[$key]->tagsBefore;
                $tag .= "<br/>".$witTagged[$key]->tags;
                $tag .= "<br/>".$witTagged[$key]->tagsAfter;
              }
            }
            $outText .= "<span title=\"$key\">".$wit[$key]."$tag</span>";
          } else {
          }
        }
        $outText .= "&nbsp;</td>\n";
      }
      $outText .= "</tr>\n";
    }
    $outText .= "</table>\n";
    return $outText;
  }
  function showXML($tagged=False) {
    $outText = "";
    $witAry = $this->getTexts();
    $matchAry = $this->getAlignment();
    $lastAry = array();
    for ($w=0;$w<count($witAry);$w++) {
      $witName = $witAry[$w]->getId();
      $last = end($witAry[$w]->getTokens());
      $lastAry[$witName] = $last->id;
    }
    $count = count($witAry);
    $total = count($witAry);
    $line = 1;
    
    $holdId = "";
    $outText = "";
    $outText = "<line n=\"1\">";
    for ($m=0;$m<count($matchAry);$m++) {
      $strAry = array();
      $idAry = array();
      $nameAry = array();
      $matches = $matchAry[$m];
      $keys = explode(",",$matches);
      $tagAry = array();
      $tagStrAry = array();
      $tagIdAry = array();
      $tagsAry = array();
      $tagsIdAry = array();
      for ($w=0;$w<count($witAry);$w++) {
        $witName = $witAry[$w]->getId();
        $wit = $witAry[$w]->getTokenAry();
        $tagWit = $witAry[$w]->getTaggedAry();
        $nameAry[] = $witName;
        foreach ($keys as $key) {
          $str = "";
          if (isset($wit[$key])) { 
            if (array_search($witName,$nameAry) !== false) {
              unset($nameAry[array_search($witName,$nameAry)]);
            }
            $str = $wit[$key]; 
            $tag = $tagWit[$key]->tags; 
            $tagStr = $tagWit[$key]->taggedWord; 
            $s = array_search($str,$strAry);
            $t = array_search($tagStr,$tagStrAry);
            if ($s !== false) {
              $idAry[$s] .= " #$witName";
            } else {
              $strAry[] = $str; $idAry[] = "#$witName";
            }
            if ($t !== false) {
              $tagIdAry[$t] .= " #$witName";
            } else {
              $tagAry[] = $tag; $tagStrAry[] = $tagStr; $tagIdAry[] = "#$witName";
            }
            if ($lastAry[$witName] == $key) {
              $count--;
              $lastAry[$witName] = "FOUND";
            }
            $tags = explode(';',$tagWit[$key]->tags); 
            foreach ($tags as $tag) {
              if (strpos($tag,':') > 0) { 
                $code = substr($tag,0,1); 
                $tag = substr($tag,2); 
                if (array_key_exists($code,$tagsAry) === false) { 
                  $tagsAry[$code] = array();
                  $tagsIdAry[$code] = array();
                }
                $t = array_search($tag,$tagsAry[$code]);
                if ($t !== false) {
                  $tagsIdAry[$code][$t] .= " #$witName";
                } else {
                  $tagsAry[$code][] = $tag; $tagsIdAry[$code][] = "#$witName";
                }
              }
            }
          } elseif ($lastAry[$witName] == "FOUND") {
              if (array_search($witName,$nameAry) !== false) {
                unset($nameAry[array_search($witName,$nameAry)]);
              }
          } else {
          }
        }
      }
      if (count($nameAry)) {
        $strAry[] = "";
        $names = "";
        foreach ($nameAry as $name) {
          if ($names) { $names .= " "; }
          $names .= "#$name";
        }
        $idAry[] = $names;
        $tagIdAry[] = $names;
        $tagAry[] = "";
        $tagStrAry[] = "";
      }
    
      $wits = 0;
      $lb = array_search("\\n",$strAry);
      if ($lb !== false) {
        $wits = substr_count($idAry[$lb],'#');
        if (($wits * 2) > $count) {
          $line++;
          $outText .= "</line><line n=\"$line\">";
          $holdId = "";
        }
      } else {
        $outText .= "<app>";
        if ($tagged !== True) {
          for ($s=0;$s<count($strAry);$s++) {
            $outText .= "<rdg resp=\"".$idAry[$s]."\">".$strAry[$s]."</rdg>";
          }
        } else {
          //initial tags
          if (array_key_exists('i',$tagsAry)) {
            for ($t=0;$t<count($tagsAry['i']);$t++) {
              $outText .= self::addXMLLine('i',$tagsIdAry['i'][$t],$tagsAry['i'][$t],0,$m);
            }
          }
          if (array_key_exists('y',$tagsAry)) {
            for ($t=0;$t<count($tagsAry['y']);$t++) {
              $outText .= self::addXMLLine('i',$tagsIdAry['y'][$t],$tagsAry['y'][$t],0,$m);
            }
          }

          //words
          for ($t=0;$t<count($tagStrAry);$t++) {
            $outText .= self::addXMLLine('word',$tagIdAry[$t],$tagStrAry[$t]);
          }

          //closing tags
          if (array_key_exists('y',$tagsAry)) {
            for ($t=0;$t<count($tagsAry['y']);$t++) {
              $outText .= self::addXMLLine('f',$tagsIdAry['y'][$t],$tagsAry['y'][$t],0,$m);
            }
          }
          if (array_key_exists('f',$tagsAry)) {
            for ($t=0;$t<count($tagsAry['f']);$t++) {
              $outText .= self::addXMLLine('f',$tagsIdAry['f'][$t],$tagsAry['f'][$t],0,$m);
            }
          }
        }
        $outText .= "</app>";
      }
    }
    $outText .= "</line>\n";

    //Loop through tags, recording start and end points, checking for mismatches
    $rdgAry = array();
    $rdgRespAry = array();
    $tagSeqAry = array();
    $rdgSeqAry = array();
    preg_match_all("/<rdg type=\"tag\" n=\"(.)\" varSeq=\"(\d+)\" ([a-z\:]+)=\"(.*?)\" resp=\"(.*?)\">(.*?)<\/rdg>/",$outText,$matchAry, PREG_SET_ORDER);
    foreach ((array)$matchAry as $match) {
      $rdg   = $match[0];
      $type  = $match[1];
      $seq   = $match[2];
      $id    = $match[4];
      $resp  = $match[5];
      $tag   = $match[6];
      if (isset($tagSeqAry[$id])) {
      //tag with the same name already exists
        if (substr($id,0,1) == "#") {
          //two end points for a single start tag; need to split start tags
          $initialId = substr($id,1);
          $initialRdg = $rdgAry[$initialId];
          $initialResp = $rdgRespAry[$id];
          //update id for current reading
          $newId = "$initialId.1";
          $newCurrentRdg = str_replace($id,"$id.1",$rdg);
          $outText = str_replace($rdg,$newCurrentRdg,$outText);
          $id .= ".1";

          //create new start tag
          $newRdg = str_replace($initialId,$newId,$initialRdg);
          $newRdg = preg_replace("/resp=\".*?\"/","resp=\"$resp\"",$newRdg);
          //add new start tag to appropriate arrays
          $rdgAry[$newId] = $newRdg;
          $rdgRespAry[$newId] = $resp;
          $initialSeq = $tagSeqAry[$initialId];
          $tagSeqAry[$newId] = $tagSeqAry[$initialId];
          $rdgSeqAry["i-$initialSeq"][] = $newId;

          //update initial start tag
          $rdgAry[$initialId] = preg_replace('/resp=\".*?\"/',"resp=\"$initialResp\"",$initialRdg);
          //append old start tag, updating the @resp from the closing tag, and update the text
          $newRdg .= $rdgAry[$initialId];
          $outText = str_replace($initialRdg,$newRdg,$outText);
        }
      }
      $rdgAry[$id] = $rdg;
      $rdgRespAry[$id] = $resp;
      $tagSeqAry[$id] = $seq;
      $key = "$type-$seq";
      if (isset($rdgSeqAry[$key])) {
        $rdgSeqAry[$key][] = $id;
      } else {
        $rdgSeqAry[$key] = array($id);
      }
    }
    // where multiple tags are in the same position, sort them by starting/ending position and tag name
    foreach ($rdgSeqAry as $key=>$idAry) {
      if (count($idAry) > 1) {
        $type = substr($key,0,1);
        $seq = substr($key,2);
        $rdgSort = array();
        foreach ($idAry as $id) {
          if ($type == "i") { 
            $sort =  substr("000000000".$tagSeqAry["#".$id],-9);
          } else if ($type == "f") { 
            $id = substr($id,1); 
            $sort =  substr("000000000".$tagSeqAry[$id],-9);
          } else {
            $sort = "";
          }
          $sort .= substr($id,0,strrpos($id,'-'));
          $rdgSort[$id] = $sort;
        }
        if ($type == "i") {
          arsort($rdgSort);
        } else {
          uasort($rdgSort,'tagSort');
        }
        $outText = preg_replace("/(<rdg type=\"tag\" n=\"$type\" varSeq=\"$seq\".*?<\/rdg>)+/","[DROMIOREPLACE]",$outText);
        $newRdg = "";
        foreach ($rdgSort as $id=>$key) {
          if ($type == "f") { $id = "#$id"; }
          $newRdg .= $rdgAry[$id];
        }
        $outText = str_replace("[DROMIOREPLACE]",$newRdg,$outText);
      }
    }
    return $outText;
  }
  function addXMLLine($type,$id,$tag,$n=0,$varSeq=0) {
    static $tagsSequence = 0;
    static $tagsSequenceAry = array();
    $text = "";
    $respAry = explode(" ",$id);
    if ($type == "i") {
      $x = $tagsSequence++;
      $xmlId = "xml:id=\"$tag\_$x\"";
      foreach ($respAry as $resp) {
        $resp = substr($resp,1);
        if (array_key_exists($resp,$tagsSequenceAry) === false) { $tagsSequenceAry[$resp] = array(); }
        $tagsSequenceAry[$resp][$tag] = $x;
      }
      $text = "<rdg type=\"tag\" n=\"i\" varSeq=\"$varSeq\" xml:id=\"$tag-$x\" resp=\"$id\">$tag</rdg>";
    } elseif ($type == "f") {
      $lineAry = array();
      foreach ($respAry as $resp) {
        $resp = substr($resp,1);
        if (array_key_exists($resp,$tagsSequenceAry)) {
          if (array_key_exists($tag,$tagsSequenceAry[$resp])) {
            $x = $tagsSequenceAry[$resp][$tag];
            unset($tagsSequenceAry[$resp][$tag]);
            if (array_key_exists($x,$lineAry)) {
              $lineAry[$x] .= " #$resp";
            } else {
              $lineAry[$x] = "#$resp";
            }
          }
        }
      }
      foreach ($lineAry as $x=>$id) {
        $text .= "<rdg type=\"tag\" n=\"f\" varSeq=\"$varSeq\" corresp=\"#$tag-$x\" resp=\"$id\">$tag</rdg>";
      }
    } elseif ($type == "word") {
      if (preg_match('/^(?P<before>.*?)\<\/f:(?P<tag>.*?)\>(?P<after>.*)$/',$tag,$match)) {
        $before = $match['before'];
        $tag = $match['tag'];
        $after = $match['after'];
        $text = self::addXMLLine('word',$id,$before);
        $text .= self::addXMLLine('f',$id,$tag,$n);
        $n++;
        $text .= self::addXMLLine('word',$id,$after,$n);
      } else {
        $nStr = "";
        if ($n > 0) { $nStr = "n=\"$n\""; }
        $text = "<rdg type=\"word\" $nStr resp=\"$id\">$tag</rdg>";
      }
    }
    return $text;
  }
}
function tagSort($a,$b) {
  $a1 = substr($a,0,9);
  $a2 = substr($a,9);
  $b1 = substr($b,0,9);
  $b2 = substr($b,9);
  if ($a1 == $b1) { return ($a2 < $b2) ? -1 : 1; }
  return ($a1 < $b1) ? 1 : -1;
}
?>
