<?php

if ($transcription_dir) {
  define("DROMIO_BASE_DIR",$transcription_dir);
} else {
  define("DROMIO_BASE_DIR","transcriptions");
}
if ($failsafe_dir) {
  define("DROMIO_FAILSAFE_DIR",$failsafe_dir);
} else {
  //define("DROMIO_FAILSAFE_DIR","failsafe");
  define("DROMIO_FAILSAFE_DIR","");
}
function hideErrors($errno,$errstr,$errfile,$errline) {
  //return true;
}
class FileNotFound extends Exception {}
class DocumentNotValid extends Exception {}
class DromioDocument extends DOMDocument {
  private $dromioVersion = "0.9.1a";
  private $base_dir = DROMIO_BASE_DIR;
  private $failsafe_dir = DROMIO_FAILSAFE_DIR;
  public $teiNS = "http://www.tei-c.org/ns/1.0";
  private $filename = "";
  private $fileHandle = "";
  private $warningMsg = "";
  private $errorMsg = "";

  function __construct($filename,$creationFlag = false) {
    parent::__construct('1.0','utf-8');
    if (!$filename) { return true; }
    $this->fn = $filename;
    $this->filename = $this->base_dir."/$filename";
    try { 
      self::loadDoc();
    } catch (FileNotFound $e) {
      if ($creationFlag == false) {
        exit("File Not Found ".$this->filename);
      } else {
      try {
        self::createDoc();
      } catch (Exception $e) {
      echo "<p>ERROR!<br/>".htmlspecialchars($this->saveXML())."</p>";
        //var_dump($e->getMessage());
        return false;
      }
      }
    } catch (DocumentNotValid $e) {
      exit($e->getMessage()." ".$this->filename);
    } catch (Exception $e) {
      var_dump($e->getMessage());
      return false;
    }
  }
  private function loadDoc() {
    if (file_exists($this->filename) === false) {
      throw new FileNotFound ('File not found.');
    }
    self::validateXMLDocument($this->filename,true);
    $this->preserveWhiteSpace = false;
    $this->load($this->filename) or exit("Unable to load file ".$this->filename);
    $this->formatOutput = true;
    self::updateVersion();
  }
  private function validateXMLDocument($doc,$isFile=true) {
    libxml_use_internal_errors(true);
    if ($isFile) {
      simplexml_load_file($doc);
    } else {
      simplexml_load_string($doc);
    }
    $errors = libxml_get_errors();
    $msg = "";
    foreach ($errors as $error) {
      if ($error->level == LIBXML_ERR_WARNING) {
        $msg .= "Warning";
      } elseif ($error->level == LIBXML_ERR_ERROR) {
        $msg .= "Error";
      } elseif ($error->level == LIBXML_ERR_FATAL) {
        $msg .= "ERROR";
      }
      $msg .= $error->code." : ".$error->message." (".$error->line.":".$error->column.")\n";
    }
    libxml_clear_errors();
    if ($msg) {
      throw new DocumentNotValid ($msg);
      return false;
    }
  }
  private function createDoc() {
    #list($dir,$id) = self::parseFilename();
    list($dir,$id) = self::_splitLast($this->filename);
    if ($dir) { self::createDir($dir); }
    $this->root = $this->createElementNS($this->teiNS,'TEI');
    $this->appendChild($this->root);
    self::createDromioElementNS($this->teiNS,'TEI:teiHeader/TEI:fileDesc/TEI:titleStmt/TEI:title','title');
    self::createDromioElementNS($this->teiNS,'TEI:text/TEI:body/p');
    $x = self::createDromioElementNS($this->teiNS,'TEI:teiHeader/TEI:encodingDesc/TEI:variantEncoding');
    $x->setAttribute('method','parallel-segmentation');
    $x->setAttribute('location','internal');
    $x = self::createDromioElementNS($this->teiNS,'TEI:teiHeader/TEI:encodingDesc/TEI:appInfo/TEI:application');
    $x->setAttribute('ident','Dromio');
    $x->setAttribute('version',$this->dromioVersion);
    self::createDromioElementNS($this->teiNS,'TEI:teiHeader/TEI:encodingDesc/TEI:appInfo/TEI:application/TEI:label','Dromio Transcription/Collation tool');
    self::writeDoc();
  }
  private function createDromioElementNS($ns,$element,$value=null,$idx=0) {
    $nodeList = self::getElementsByNodePathNS($ns, $element);
    if ($idx == -1 || $nodeList->length <= $idx) {
      list($parent,$new) = self::_splitLast($element);
      list($new,$attributes) = self::_splitAttributes($new);
      $new = str_replace('TEI:','',$new);
      if ($parent) {
        $pNode = self::createDromioElementNS($ns,$parent);
      } else {
        $pNode = $this->documentElement;
      }
      if ($pNode) {
        $x = $this->createElementNS($this->teiNS,$new,$value);
        foreach ($attributes as $att=>$val) {
          $x->setAttribute($att,$val);
        }
        if ($idx == -1) {
          $pNode->insertBefore($x,$pNode->firstChild);
        } else {
          $pNode->appendChild($x);
        }
        return $x;
      }
    } else {
      return $nodeList->item($idx);
    }
  }
  private function removeDromioElementNS($ns,$element) {
    $nodeList = self::getElementsByNodePathNS($ns, $element);
    if ($nodeList->length == 0) {
      list($parent,$new) = self::_splitLast($element);
      $new = str_replace('TEI:','',$new);
      if ($parent) {
        $pNode = self::createDromioElementNS($ns,$parent);
        if ($pNode) {
          $x = $this->createElementNS($this->teiNS,$new,$value);
          $pNode->appendChild($x);
          return $x;
        }
      } else {
        $x = $this->createElementNS($this->teiNS,$new,$value);
        $pNode = $this->root;
        $pNode->appendChild($x);
        return $x;
      }
    } elseif ($nodeList->length == 1) {
      return $nodeList->item(0);
    }
  }
  public function getElementsByNodePathNS($ns, $path, $prefix='TEI') {
    $xpath = new DromioXpath($this);
    set_error_handler("hideErrors",E_WARNING);
    $list = $xpath->query($path);
    restore_error_handler();
    if ($list === false) {
      echo "<p style='color:red;'>XPath error: $path</p>";
      return false;
    } else {
      return $list;
    }
  }
  private function validateDocId() {
  }
  private function createDir($dir) {
    $dirs = explode("/",$dir);
    $dirx = "";
    foreach ($dirs as $dir) {
      if ($dirx) { $dirx .= "/$dir"; } else { $dirx = $dir; }
      if (file_exists($dirx) && is_dir($dirx)) {
      } else { mkdir($dirx); }
    }
  }
  private function _splitLast($str,$delim="/") {
    if (strrpos($str,$delim) > 0) {
      $str1 = substr($str,0,strrpos($str,$delim));
      $str2 = substr($str,strrpos($str,$delim)+1);
      return array($str1,$str2);
    } else {
      return array("",$str);
    }
  }
  private function _splitAttributes($element,$delim="[") {
    $attributes = array();
    if (strpos($element,$delim) > 0) {
      $ary = explode($delim,$element);
      $element = array_shift($ary);
      foreach ($ary as $att) {
        if (preg_match('/@([a-zA-Z0-9:]*)=(.*?)\]/',$att,$matches)) {
          $key = $matches[1];
          $val = $matches[2];
          $val = preg_replace('/^[\'\"]/','',$val);
          $val = preg_replace('/[\'\"]$/','',$val);
          $attributes[$key] = $val;
        }
      }
    } else {
    }
    return array($element,$attributes);
  }
  private function parseFilename() {
    if (strrpos($this->filename,'/') > 0) {
      $dir = substr($this->filename,0,strrpos($this->filename,'/'));
      $file = substr($this->filename,strrpos($this->filename,'/')+1);
      return array($dir,$file);
    } else {
      return array("",$this->filename);
    }
  }
  private function validateXMLId($str,$nd) {
    if (strlen($str) == 0) {
      $this->appendMsg('error',"XML:ID is required<br/>");
      return false;
    }
    $invalidStart = preg_replace('/[a-zA-Z\:\_]/','',substr($str,0,1));
    if (strlen($invalidStart) > 0) {
      $this->appendMsg('error',"XML:ID begins with an invalid character ($invalidStart).<br/>");
      return false;
    }
    $invalidChars = preg_replace('/[a-zA-Z0-9\:\-\_\.]/','',$str);
    if (strlen($invalidChars) > 0) {
      $this->appendMsg('error',"XML:ID contains invalid characters ($invalidChars).<br/>");
      return false;
    }
    //test for uniqueness
    $nodeList = self::getElementsByNodePathNS($this->teiNS, "//TEI:*[@xml:id='$str']");
    $cnt = 0;
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) {
        if (!$node->isSameNode($nd)) { $cnt++; }
      }
      if ($cnt) {
        $this->appendMsg('error',"XML:ID already in use.<br/>");
        return false;
      }
    }
    return true;
  }
  public function updateVersion() {
    $version = $this->dromioVersion;
    $versionNodes = self::getElementsByNodePathNS($this->teiNS, '//TEI:teiHeader/TEI:encodingDesc/TEI:appInfo/TEI:application');
    if ($versionNodes->length > 0) { 
      $vNode = $versionNodes->item(0); 
      $versionOld = $vNode->getAttribute('version');
    }
    if ($versionOld < "0.5.0") {
      $sourceAry = array();
      $sourceAry = $this->getSources();
      if (count($sourceAry)) {
      foreach ($sourceAry as $sid => $type) {
        $surfaceAry = array();
        $surfaceAry = $this->getSurfaces($sid);
        if (count($surfaceAry)) {
        foreach ($surfaceAry as $sfcid => $n) {
          $respAry = array();
          $respAry = $this->getSurfaceTranscribers($sfcid,'rdg');
          if (count($respAry)) {
          foreach ($respAry as $resp) {
            $lines = $this->getDromioLines($sfcid,$resp);
            $lineAry = explode("<br/>",$lines);
            $this->setDromioLines($sfcid,$resp,$lineAry);
          }}
        }}
      }}
      $nodeList = self::getElementsByNodePathNS($this->teiNS, "//TEI:line");
      if ($nodeList->length > 0) {
        foreach ($nodeList as $node) {
          self::removeDromioNode($node,true);
        }
      }
      $nodeList = self::getElementsByNodePathNS($this->teiNS, "//TEI:textlang");
      if ($nodeList->length > 0) {
        foreach ($nodeList as $node) {
          $newNode = $this->createElementNS($this->teiNS,'textLang');
          $newNode->nodeValue = $node->nodeValue;
          $node->parentNode->replaceChild($newNode,$node);
        }
      }
      $nodeList = self::getElementsByNodePathNS($this->teiNS, "//TEI:notesStmt/TEI:p");
      if ($nodeList->length > 0) {
        foreach ($nodeList as $node) {
          $newNode = $this->createElementNS($this->teiNS,'note');
          $newNode->nodeValue = $node->nodeValue;
          $node->parentNode->replaceChild($newNode,$node);
        }
      }
      $node =  self::createDromioElementNS($this->teiNS,"TEI:teiHeader/TEI:revisionDesc/TEI:change","",-1);
      $node->setAttribute("when",date('c'));
      $node->nodeValue = "Updated to version $version"; 
      if ($vNode) { $vNode->setAttribute('version',$version); }
      self::validateXML();
      self::writeDoc();
    }
  }
  public function validateXML() {
//    self::validateMedia();
    //try to put fields in proper order for TEI validation
    self::moveNode("//TEI:fileDesc/TEI:titleStmt");
    self::moveNode("//TEI:fileDesc/TEI:editionStmt");
    self::moveNode("//TEI:fileDesc/TEI:extent");
    self::moveNode("//TEI:fileDesc/TEI:publicationStmt");
    self::moveNode("//TEI:fileDesc/TEI:publicationStmt/TEI:availability");
    self::moveNode("//TEI:fileDesc/TEI:seriesStmt");
    self::moveNode("//TEI:fileDesc/TEI:notesStmt");
    self::moveNode("//TEI:fileDesc/TEI:sourceDesc");
    self::moveNode("//TEI:msDesc/TEI:msIdentifier");
    self::moveNode("//TEI:msDesc/TEI:msContents");
    self::moveNode("//TEI:msDesc/TEI:physDesc");
    self::moveNode("//TEI:msDesc/TEI:history");
    self::moveNode("//TEI:msDesc/TEI:additional");
    self::moveNode("//TEI:msDesc/TEI:msPart");
    self::moveNode("//TEI:msIdentifier/TEI:repository");
    self::moveNode("//TEI:msIdentifier/TEI:collection");
    self::moveNode("//TEI:msIdentifier/TEI:idno");
    self::moveNode("//TEI:msIdentifier/TEI:msName");
    self::moveNode("//TEI:msItem/TEI:locus");
    self::moveNode("//TEI:msItem/TEI:textLang");
    self::moveNode("//TEI:msItem/TEI:author");
    self::moveNode("//TEI:msItem/TEI:editor");
    self::moveNode("//TEI:msItem/TEI:title");
    self::moveNode("//TEI:msItem/TEI:incipit");
    self::moveNode("//TEI:msItem/TEI:note");
    self::validateMoveText();
  }
  public function validateMedia($path) {
    if (isset($path)) {
      $nodeList = self::getElementsByNodePathNS($this->teiNS, $path);
    } else {
      $nodeList = self::getElementsByNodePathNS($this->teiNS, "//TEI:media");
    }
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) {
        $url = $node->getAttribute("url"); 
        if (strlen($url) == 0) {
          self::removeDromioNode($node,true);
        } else {
          if(function_exists('finfo_file')) {
            $mediaFile = file_get_contents($url);
            if($mediaFile) {
              $tempFile = tempnam(sys_get_temp_dir(), 'MIMEtest');
              $tmp = fopen($tempFile,"w");
              fwrite($tmp,$mediaFile);
              fclose($tmp);
              $finfo = finfo_open(FILEINFO_MIME_TYPE);
              $mime = finfo_file($finfo, $tempFile);
              $node->setAttribute("mimeType",$mime);
              unlink($tempFile);
            }
          }
        }
      }
    }
  }
  private function validateMoveText() {
    $nodeList = self::getElementsByNodePathNS($this->teiNS, "TEI:text");
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) {
        $moveNode = $this->importNode($node,true);
        $this->documentElement->appendChild($moveNode);
      }
    }
  }
  private function moveNode($path) {
    $nodeList = self::getElementsByNodePathNS($this->teiNS, $path);
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) {
        $parentNode = $node->parentNode;
        $moveNode = $this->importNode($node,true);
        $parentNode->appendChild($moveNode);
      }
    }
  }
  public function moveSurface($sfcid) {
    $path = "//TEI:surface[@xml:id='$sfcid']";
    self::moveNode($path);
  }
  public function removeSurface($sfcid) {
    $path = "//TEI:surface[@xml:id='$sfcid']";
    $nodeList = self::getElementsByNodePathNS($this->teiNS, $path);
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) {
        self::removeDromioNode($node,true);
      }
    }
  }
  public function validateSchema() {
    return true;
  }
  public function getMsg($messageType) {
    $msg = $messageType."Msg";
    if (isset($this->$msg)) {
      return $this->$msg;
    }
    return false;
  }
  public function clearMsg($messageType) {
    $msg = $messageType."Msg";
    if (isset($this->$msg) && substr($msg,-3) == "Msg") {
      $this->$msg = "";
      return true;
    }
    return false;
  }
  public function appendMsg($messageType,$str) {
    $msg = $messageType."Msg";
    if (isset($this->$msg)) {
      $this->$msg .= "$str\n";
      return true;
    }
    return false;
  }
  public function writeDoc() {
    if (strlen($this->getMsg('error')) > 0) {
        error_log($this->getMsg('error'));
    } elseif ($this->fileHandle) {
      $text = $this->saveXML();
      if ($text === false) { return false; }

      list($dir,$id) = self::_splitLast($this->filename);
      if (file_exists($this->filename) && !empty($this->failsafe_dir)) {
        copy($this->filename,"$this->failsafe_dir/".$id.".".date("YmdHis").substr((string)microtime(),2,6));
      }
      try {
        self::validateXML();
      } catch (Exception $e) {
        echo($e->getMessage()." ".$this->filename);
        return false;
      }
      try {
        self::validateXMLDocument($text,false);
        ftruncate($this->fileHandle,0);
        rewind($this->fileHandle);
        fwrite($this->fileHandle,$text);
        return true;
      } catch (Exception $e) {
        echo($e->getMessage()." ".$this->filename);
      }
      return false;
    } else {
      self::lockDoc();
      $test = self::writeDoc();
      self::unlockDoc();
      return $test;
    }
    return false;
  }
  public function lockDoc() {
    $fp = fopen($this->filename,'c+');
    if ($fp) {
      $i = 0;
      while (!flock($fp, LOCK_EX)) { 
        if ($i > 30) { return false; }
        $i++;
        usleep(100000); 
      }
      $this->fileHandle = $fp;
    }
    return $fp;
  }
  public function unlockDoc() {
    if ($this->fileHandle) {
      fclose($this->fileHandle);
      $this->fileHandle = "";
    }
  }
  function deleteDoc() {
    if (file_exists($this->filename)) {
      unlink($this->filename);
    }
  }
  function getTitle() {
    $xpath = new DromioXpath($this);
    $nodes = $xpath->query('TEI:teiHeader/TEI:fileDesc/TEI:titleStmt/TEI:title');
    if ($nodes->length > 0) {
      return $nodes->item(0)->nodeValue;
    } else {
      return '[No title]';
    }
  }
  function getURL() {
    return $this->filename;
  }
  function getSources() {
    $rtnAry = array();
    foreach ($GLOBALS['levelAry'] as $level) {
      foreach ($GLOBALS[$level.'Ary'] as $node) {
        if (isset($node->id) && isset($GLOBALS['sourceIdAry'][$node->id])) {
          $type = $GLOBALS['sourceIdAry'][$node->id];
          $path = $node->path;
          $path = preg_replace('/\[.*?\]/','',$path);
          $values = self::getNodeValue($path);
          foreach ($values as $id) {
            $rtnAry[$id] = $type;
          }
        }
      }
    }
    return $rtnAry;
  }
  function getSurfaces($sourceId=null) {
    $rtnAry = array();
    if ($sourceId) {
      $path = "//TEI:sourceDoc[@corresp='$sourceId']/TEI:surface";
    } else {
      $path = "//TEI:surface";
    }
    $nodeList = self::getElementsByNodePathNS($this->teiNS, $path);
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) { 
        $id = $node->getAttribute("xml:id"); 
        $n = $node->getAttribute("n"); 
        $rtnAry[$id] = $n; 
      }
    }
    return $rtnAry;
  }
  function getSurfaceTranscribers($surfaceId=null,$element="zone") {
    $rtnAry = array();
    if ($surfaceId) {
      $path = "//TEI:surface[@xml:id='$surfaceId']//TEI:$element";
    } else {
      $path = "//TEI:$element";
    }
    $nodeList = self::getElementsByNodePathNS($this->teiNS, $path);
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) { 
        $resp = $node->getAttribute("resp"); 
        if ($resp) {
          $resp = str_replace("#","",$resp);
          $rtnAry[$resp] = $resp; 
        }
        $wit = $node->getAttribute("wit"); 
        if ($wit) {
          $wit = str_replace("#","",$wit);
          $rtnAry[$wit] = $wit;
        }
      }
    }
    return $rtnAry;
  }
  function getValue($path) {
    if (!$path) { return false; }
    list($node,$attribute) = self::_splitLast($path);
    if (substr($attribute,0,1) == "@") {
      $attribute = substr($attribute,1);
      return self::getAttributeValue($node,$attribute);
    } else {
      return self::getNodeValue($path);
    }
  }
  function getNodeValue($path) {
    $rtn = array();
    if (!$path) { return false; }
    $nodeList = self::getElementsByNodePathNS($this->teiNS, $path);
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) { $rtn[] = $node->nodeValue; }
    }
    return $rtn;
  }
  function getAttributeValue($path,$attribute) {
    $rtn = array();
    if (!$path) { return false; }
    $nodeList = self::getElementsByNodePathNS($this->teiNS, $path);
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) { $rtn[] = $node->getAttribute($attribute); }
    }
    return $rtn;
  }
  function removeDromioNode($node,$recursive=false) {
    if ($node) {
      $pNode = $node->parentNode;
    }
    if ($pNode) {
    $pNode->removeChild($node);
    if ($recursive && $pNode->hasChildNodes() == false) {
      self::removeDromioNode($pNode,true);
    }
    }
  }
  function setValue($path,$value,$idx=0,$xmlFlag=false) {
    if (!$path) { return false; }
    list($node,$attribute) = self::_splitLast($path);
    if (substr($attribute,0,1) == "@") {
      $attribute = substr($attribute,1);
      self::setAttributeValue($node,$attribute,$value,$idx);
    } else {
      self::setNodeValue($path,$value,$idx,$xmlFlag);
    }
    return true;
  }
  function setNodeValue($path,$value,$idx=0,$xmlFlag=false) {
    if (!$path) { return false; }
    $nodeList = self::getElementsByNodePathNS($this->teiNS, $path);
    $node = $nodeList->item($idx);
    if (strlen($value) == 0) {
      if ($nodeList->length > $idx) {
        self::removeDromioNode($nodeList->item($idx),true);
      }
    } elseif ($nodeList->length > $idx) {
      $nodeList->item($idx)->nodeValue = $value; 
    } else {
      $node = self::createDromioElementNS($this->teiNS,$path,$value,$idx);
    }
    if ($xmlFlag && $value) { self::setNodeValueXML($node,$value); }
    return true;
  }
  function setNodeValueXML($node,$xmlStr) {
    try {
      $validateCheck = new DOMDocument;
      $validateCheck->loadXML("<xml>$xmlStr</xml>");
      $node->nodeValue = "";
      $xmlFragment = $this->createDocumentFragment();
      if ($xmlFragment->appendXML($xmlStr)) {
        $node->appendChild($xmlFragment);
      }
    } catch (Exception $e) {
      exit("Error adding XML string to ".$this->filename.": $xmlStr");
    }
    return true;
  }
  function setAttributeValue($path,$att,$value,$idx=0) {
    if (!$path) { return false; }
    $node = self::createDromioElementNS($this->teiNS,$path,"",$idx);
    if (strlen($value) == 0) {
      $node->removeAttribute($att);
    } elseif ($att == "xml:id") {
      if ($this->validateXMLId($value,$node)) {
        $node->setAttribute($att,$value);
      }
    } else {
      $node->setAttribute($att,$value);
    }
    return true;
  }

  function getDromioLines($sfcid,$witid) {
    $lineStr = "";
    $nodeList = self::getElementsByNodePathNS($this->teiNS, "//TEI:surface[@xml:id='$sfcid']/TEI:zone[@resp='#$witid']");
    $nodeListOld = self::getElementsByNodePathNS($this->teiNS, "//TEI:surface[@xml:id='$sfcid']/TEI:line/TEI:app/TEI:rdg[@resp='#$witid']");
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) {
        $lineStr .= $node->nodeValue."<br/>";
      }
    } elseif ($nodeListOld->length > 0) {
      foreach ($nodeListOld as $node) {
        $lineStr .= $node->nodeValue."<br/>";
      }
    }
    $lineStr = str_replace("<lb/>","<br/>",$lineStr);
    $lineStr = str_replace("'","&#39;",$lineStr);
    $lineStr = str_replace("\"","&#34;",$lineStr);
    $lineStr = str_replace(" ","&nbsp;",$lineStr);
    $lineStr = str_replace("&lt;","<",$lineStr);
    $lineStr = str_replace("&gt;",">",$lineStr);
    while (preg_match('/<([^>]*)\&nbsp;([^>]*)>/',$lineStr)) {
      $lineStr = preg_replace('/<([^>]*)\&nbsp;([^>]*)>/',"<$1 $2>",$lineStr);
    }
    while (preg_match('/<([^>]*)\&#34;([^>]*)>/',$lineStr)) {
      $lineStr = preg_replace('/<([^>]*)\&#34;([^>]*)>/',"<$1\"$2>",$lineStr);
    }
    
    return $lineStr;
  }
  function getDromioTranscription($sfcid,$witid) {
    $lineStr = "";
    $nodeList = self::getElementsByNodePathNS($this->teiNS, "//TEI:surface[@xml:id='$sfcid']/TEI:zone[@resp='#$witid']");
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) {
        foreach ($node->childNodes as $child) {
          $lineStr .= $this->saveXML($child);
        }
      }
    }
    return $lineStr;
  }
  function addDromioWitness($witid) {
    //add witness if doesn't exist
    $nodeList = self::getElementsByNodePathNS($this->teiNS, "//*[@xml:id='$witid']");
    if ($nodeList->length == 0) {
      self::setValue("TEI:teiHeader/TEI:fileDesc/TEI:titleStmt/TEI:respStmt[@xml:id='$witid']/TEI:name","$witid");
      self::setValue("TEI:teiHeader/TEI:fileDesc/TEI:titleStmt/TEI:respStmt[@xml:id='$witid']/TEI:resp",'Transcriber');
    }
  }
  function setDromioTranscription($sfcid,$witid,$lineStr) {
    self::addDromioWitness($witid);
    
    $lineStr = preg_replace('/(<lb\/>)*\s*$/',"",$lineStr);
    self::setValue("//TEI:surface[@xml:id='$sfcid']/TEI:zone[@resp='#$witid']",$lineStr,0,true);
    return true;
  }

  function setDromioLines($sfcid,$witid,$lineAry) {
    //an older way to handle Dromio transcriptions, supercededy by setDromioTranscription
    //add witness if doesn't exist
    $nodeList = self::getElementsByNodePathNS($this->teiNS, "//*[@xml:id='$witid']");
    if ($nodeList->length == 0) {
      self::setValue("TEI:teiHeader/TEI:fileDesc/TEI:titleStmt/TEI:respStmt[@xml:id='$witid']/TEI:name","$witid");
      self::setValue("TEI:teiHeader/TEI:fileDesc/TEI:titleStmt/TEI:respStmt[@xml:id='$witid']/TEI:resp",'Transcriber');
    }
    
    //add new lines
    $lineStr = join('<lb/>',$lineAry);
    $lineStr = str_replace("<br/>","<lb/>",$lineStr);
    $lineStr = str_replace("&#39;","'",$lineStr);
    $lineStr = str_replace("&#34;","\"",$lineStr);
    $lineStr = str_replace("&nbsp;"," ",$lineStr);
    $lineStr = htmlspecialchars($lineStr, ENT_QUOTES, 'UTF-8');
    $lineStr = str_replace("<","&lt;",$lineStr);
    $lineStr = str_replace(">","&gt;",$lineStr);
    $lineStr = str_replace("&lt;lb/&gt;","<lb/>",$lineStr);
    $lineStr = preg_replace('/(<lb\/>)*\s*$/',"",$lineStr);
    $lineStr .= "<lb/>";
    self::setValue("//TEI:surface[@xml:id='$sfcid']/TEI:zone[@resp='#$witid']",$lineStr);

    return true;
  }

  private function validateMoveLines() {
    $nodeList = self::getElementsByNodePathNS($this->teiNS, "TEI:line");
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) {
        $n = $node->getAttribute('n');
        $moveNode = $this->importNode($node,true);
        $this->documentElement->appendChild($moveNode);
      }
    }
  }
  function setDromioStatus($status,$who=null,$target=null,$source=null) {
    $node =  self::createDromioElementNS($this->teiNS,"TEI:teiHeader/TEI:revisionDesc/TEI:change","",-1);
    $date = date('c');
    $node->setAttribute("when",$date);
    if ($target) { $node->setAttribute("target","#".$target); }
    $node->setAttribute("status",$status);
    if ($who) { $node->setAttribute("who","#".$who); }
    if ($source) { $node->setAttribute("source","#".$source); }
    return true;
  }
  function getDromioStatus($status=null,$who=null,$target=null) {
    $rtnAry = array();
    $xpath = new DromioXpath($this);
    $path = 'TEI:teiHeader/TEI:revisionDesc/TEI:change';
    if ($status) { $path .= "[@status='$status']"; }
    if ($target) { $path .= "[@target='#$target']"; }
    if ($who) { $path .= "[@who='#$who']"; }
    $nodeList = $xpath->query($path);
    if ($nodeList->length > 0) {
      foreach ($nodeList as $node) { 
        $status = $node->getAttribute("status"); 
        $who = $node->getAttribute("who"); 
        $who = str_replace("#","",$who);
        $target = $node->getAttribute("target"); 
        $target = str_replace("#","",$target);
        $when = $node->getAttribute("when"); 
        $rtnAry[] = array($status,$who,$target,$when); 
      }
    }
    return $rtnAry;
  }
}

class DromioXpath extends DOMXpath {
  function __construct($xmlDoc,$prefix="TEI") {
    parent::__construct($xmlDoc);
    if ($xmlDoc->teiNS) {
      $this->registerNamespace($prefix,$xmlDoc->teiNS);
    }
  }
}

?>
