class FloatingBox {
  constructor() {
    this.boxLeft = 5;
    this.boxTop = 60;
    this.boxWidth = 500;
    this.boxHeight = 180;
    this.minWidth = 200; 
    this.minHeight = 50; 
    this.frameWidth = 5;
    this.topFrameHeight = 4;
    this.btnH = 18; //button height
    this.tbH = 21; //toolbar height

    this.dragX = 0;
    this.dragY = 0;
    this.dragDirection = "";

    this.toolbarAry = [];
    this.tbAry = [];
    this.buttonAry = [];
    this.zoomFactor = 1;

    var boxDiv = document.createElement('div');
    boxDiv.className = "box noprint";
    document.body.appendChild(boxDiv);
    this.boxDiv = boxDiv;

    var boxContent = document.createElement('div');
    boxContent.className = "box noprint";
    boxContent.setAttribute('id','boxContent');
    boxContent.style.overflow = "auto";
    boxContent.style.userSelect = "none";
    boxDiv.appendChild(boxContent);
    this.boxContent = boxContent;

    var self = this;
    var frameTop = document.createElement('div');
    frameTop.className = "boxFrame";
    frameTop.draggable = "true";
    frameTop.ondragstart = function(event) {self.dragBox(event);};
    boxDiv.appendChild(frameTop);
    this.frameTop = frameTop;

    var frameBottom = document.createElement('div');
    frameBottom.className = "boxFrame";
    frameBottom.style.cursor = "n-resize";
    frameBottom.draggable = "true";
    frameBottom.ondragstart = function(event) {self.resizeBox(event,'S');};
    boxDiv.appendChild(frameBottom);
    this.frameBottom = frameBottom;

    var frameLeft = document.createElement('div');
    frameLeft.className = "boxFrame";
    frameLeft.style.cursor = "e-resize";
    frameLeft.draggable = "true";
    frameLeft.ondragstart = function(event) {self.resizeBox(event,'W');};
    boxDiv.appendChild(frameLeft);
    this.frameLeft = frameLeft;

    var frameRight = document.createElement('div');
    frameRight.className = "boxFrame";
    frameRight.style.cursor = "e-resize";
    frameRight.draggable = "true";
    frameRight.ondragstart = function(event) {self.resizeBox(event,'E');};
    boxDiv.appendChild(frameRight);
    this.frameRight = frameRight;

    var frameLeftCorner = document.createElement('div');
    frameLeftCorner.className = "boxFrame";
    frameLeftCorner.style.cursor = "ne-resize";
    frameLeftCorner.draggable = "true";
    frameLeftCorner.ondragstart = function(event) {self.resizeBox(event,'SW');};
    boxDiv.appendChild(frameLeftCorner);
    this.frameLeftCorner = frameLeftCorner;

    var frameRightCorner = document.createElement('div');
    frameRightCorner.className = "boxFrame";
    frameRightCorner.style.cursor = "nw-resize";
    frameRightCorner.draggable = "true";
    frameRightCorner.ondragstart = function(event) {self.resizeBox(event,'SE');};
    boxDiv.appendChild(frameRightCorner);
    this.frameRightCorner = frameRightCorner;

    this.redrawBox();
  }
  get position() { return {left:this.boxLeft, top:this.boxTop}; }
  get size() { return {width:this.boxWidth, height:this.boxHeight}; }
  get scroll() { return {x:this.boxContent.scrollLeft, y:this.boxContent.scrollTop}; }
  set position(position) { this.boxLeft=position.left; this.boxTop=position.top; this.moveBox(); }
  set size(size) { this.boxWidth=size.width; this.boxHeight=size.height; this.redrawBox(); }
  set scroll(scroll) { this.boxContent.scrollTo(scroll.x,scroll.y); }

  addToolbar(id,display="true") {
    var tb = document.createElement('div');
    tb.setAttribute('id',id+'TB');
    tb.className = "toolbar";
    this.frameTop.appendChild(tb);
    if (display == "true") {
      this.topFrameHeight += this.tbH;
      this.minHeight += this.tbH;
    } else {
      tb.style.display = "none";
    }
    this.toolbarAry[id] = 0;
    this.redrawBox();
  }
  selectToolbar(id) {
    var tb = document.getElementById(id+'TB');
    if (tb.style.display == "none") {
      tb.style.display = "block";
      this.topFrameHeight += this.tbH;
      this.minHeight += this.tbH;
    } else {
      tb.style.display = "none";
      this.topFrameHeight -= this.tbH;
      this.minHeight -= this.tbH;
    }
    this.redrawBox();
  }
  addToolbarButton(toolbar,id,type,label,title,tag,className,style) {
    var btn = document.createElement('button');
    btn.setAttribute('id',id+'Btn');
    btn.className = "toolbarButton";
    if (type == "shortcut") { btn.className += " toolbarShortcut"; }
    if (className) { btn.className += " "+className; }
    if (style) { btn.style.cssText = style; }
    btn.innerHTML = label;
    btn.title = title;
    var self = this;
    if (type == "shortcut") {
      btn.onclick = function(tag) { return function() { self.addShortcut(tag);};}(tag);
    } else if (type == "tag") {
      btn.onclick = function(tag) { return function() { self.addTag(tag);};}(tag);
    } else if (type == "function") {
      var fn = window[tag];
      if (typeof fn === "function") { btn.onclick = function() { fn(); } }
    } else {
    }
    var tb = document.getElementById(toolbar+'TB');
    if (tb === null) { } else {
      tb.appendChild(btn);
      var t = this.toolbarAry[toolbar];
      var width = btn.offsetWidth;
      t += width;
      if (t > this.minWidth) { this.minWidth = t; }
      this.toolbarAry[toolbar] = t;
    }
  }
  addToolbarPadding(toolbar,className) {
    var btn = document.createElement('span');
    btn.style.display = "inline-block";
    btn.className = "toolbarButton";
    if (className) { btn.className += " "+className; }
    var tb = document.getElementById(toolbar+'TB');
    if (tb === null) { } else {
      tb.appendChild(btn);
      var t = this.toolbarAry[toolbar];
      var width = btn.offsetWidth;
      t += width;
      if (t > this.minWidth) { this.minWidth = t; }
      this.toolbarAry[toolbar] = t;
    }
  }
  addToolbarText(toolbar,text,id,className) {
    var btn = document.createElement('span');
    if (id) { btn.setAttribute('id',id); }
    btn.style.display = "inline-block";
    btn.className = "toolbarText";
    if (className) { btn.className += " "+className; }
    btn.textContent = text;
    var tb = document.getElementById(toolbar+'TB');
    if (tb === null) { } else {
      tb.appendChild(btn);
      var t = this.toolbarAry[toolbar];
      var width = btn.offsetWidth;
      t += width;
      if (t > this.minWidth) { this.minWidth = t; }
      this.toolbarAry[toolbar] = t;
    }
    return btn;
  }

  moveBox() {
    this.boxDiv.style.left = this.boxLeft+"px";
    this.boxDiv.style.top = this.boxTop+"px";
  }
  redrawBox() {
    let minWidth = this.minWidth + this.frameWidth + this.frameWidth; 
    if (this.boxWidth < minWidth) { this.boxWidth = minWidth; }

    var frame = this.boxContent;
    let topFrameHeight = this.topFrameHeight;
    if (topFrameHeight < this.tbH) { topFrameHeight = this.tbH; }
    frame.style.left = this.frameWidth+"px"; frame.style.top = this.topFrameHeight+"px";
    frame.style.width = (this.boxWidth-this.frameWidth-this.frameWidth)+"px"; 
    frame.style.height = (this.boxHeight-this.topFrameHeight-this.frameWidth)+"px";

    var frame = this.frameTop;
    frame.style.left = "0px"; frame.style.top = "0px";
    frame.style.width = this.boxWidth+"px"; frame.style.height = topFrameHeight+"px";

    var frame = this.frameBottom;
    frame.style.left = this.frameWidth+"px"; frame.style.top = (this.boxHeight-this.frameWidth)+"px";
    frame.style.width = (this.boxWidth-this.frameWidth-this.frameWidth)+"px"; frame.style.height = this.frameWidth+"px";

    var frame = this.frameLeft;
    frame.style.left = "0px"; frame.style.top = this.topFrameHeight+"px";
    frame.style.width = this.frameWidth+"px"; frame.style.height = (this.boxHeight-this.topFrameHeight-this.frameWidth)+"px";

    var frame = this.frameRight;
    frame.style.left = (this.boxWidth-this.frameWidth)+"px"; frame.style.top = this.topFrameHeight+"px";
    frame.style.width = this.frameWidth+"px"; frame.style.height = (this.boxHeight-this.topFrameHeight-this.frameWidth)+"px";

    var frame = this.frameLeftCorner;
    frame.style.left = "0px"; frame.style.top = (this.boxHeight-this.frameWidth)+"px";
    frame.style.width = this.frameWidth+"px"; frame.style.height = this.frameWidth+"px";

    var frame = this.frameRightCorner;
    frame.style.left = (this.boxWidth-this.frameWidth)+"px"; frame.style.top = (this.boxHeight-this.frameWidth)+"px";
    frame.style.width = this.frameWidth+"px"; frame.style.height = this.frameWidth+"px";

    frame = document.getElementById('transcriptionFrame');
    if (frame) {
        frame.style.width = (this.boxWidth-this.frameWidth-this.frameWidth)+"px"; 
        frame.style.height = (this.boxHeight-this.topFrameHeight-this.frameWidth)+"px";
        frame.style.overflow = "auto";
    }
    if (document.getElementById('textholder')) {
      document.getElementById('textholder').style.top = (this.boxHeight + this.boxTop + 30)+"px";
    }

  }

  dropBox(event) {
    this.dragX = 0;
    this.dragY = 0;
    this.dragDirection = "";
    event.preventDefault();
    return false;
  }
  dragBox(event) {
    event.dataTransfer.setData('text/html',null);
    var img = document.createElement('img');
    img.src = 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7';
    event.dataTransfer.setDragImage(img, 0, 0);
    this.dragX = event.clientX;
    this.dragY = event.clientY;
    var self = this;
    document.ondragover = function(event) { self.draggingBox(event); };
    document.ondrop = function(event) { self.dropBox(event); };
  }
  draggingBox(event) {
    this.boxLeft = parseInt(this.boxDiv.style.left) + (event.clientX - this.dragX);
    this.boxDiv.style.left = this.boxLeft + "px";
    this.boxTop = parseInt(this.boxDiv.style.top) + (event.clientY - this.dragY);
    this.boxDiv.style.top = this.boxTop + "px";
    this.dragX = event.clientX;
    this.dragY = event.clientY;
    event.preventDefault();
    return false;
  }
  resizeBox(event,direction) {
    event.dataTransfer.setData('text',null);
    var img = document.createElement('img');
    img.src = 'data:image/gif;base64,R0lGODlhAQABAIAAAAAAAP///yH5BAEAAAAALAAAAAABAAEAAAIBRAA7';
    event.dataTransfer.setDragImage(img, 0, 0);
    this.dragX = event.clientX;
    this.dragY = event.clientY;
    this.dragDirection = direction;
    var self = this;
    document.ondragover = function(event) { self.resizeBoxOver(event); };
    document.ondrop = function(event) { self.dropBox(event); };
  }
  resizeBoxOver(event) {
    var newX = false;
    var newW = 0;
    var newH = 0;
    var redraw = false;
    if (this.dragDirection && this.dragDirection.indexOf("S") > -1) {
      newH = this.boxHeight + (event.clientY - this.dragY);
      if (newH < this.minHeight) { newH = this.minHeight; }
    }
    if (this.dragDirection && this.dragDirection.indexOf("E") > -1) {
      newW = this.boxWidth + (event.clientX - this.dragX);
    }
    if (this.dragDirection && this.dragDirection.indexOf("W") > -1) {
      newX = this.boxLeft + (event.clientX - this.dragX);
      newW = this.boxWidth - (event.clientX - this.dragX);
    }
    if (newH > this.minHeight) { this.boxHeight = newH; this.boxDiv.style.height = this.boxHeight+"px"; redraw = true; }
    if (newW > this.minWidth) { this.boxWidth = newW; this.boxDiv.style.width = this.boxWidth+"px"; redraw = true; }
    if (newX !== false && newW > this.minWidth) { this.boxLeft = newX; this.boxDiv.style.left = this.boxLeft+"px"; redraw = true; }
    if (redraw) { this.redrawBox(); }
    this.dragX = event.clientX;
    this.dragY = event.clientY;
    event.preventDefault();
    return false;
  }

  selectImgWin(fobj,e) { }

}
