class FloatingTranscription extends FloatingBox {
  constructor(text) {
    super(text);
    var transcriptionFrame = document.createElement("div");
    transcriptionFrame.className = "transcriptionFrame";
    this.boxContent.appendChild(transcriptionFrame);
    transcriptionFrame.contentEditable = true;
    transcriptionFrame.innerHTML = this.cleanSelfClosingTags(text);
    //transcriptionFrame.innerHTML = text;

    var self = this;
    transcriptionFrame.addEventListener("load",function(e){this.contentWindow.document.designMode = "on"; transcriptionFrame.contentWindow.addEventListener('keyup',detectKey, true);}, false);

    this.transcriptionFrame = transcriptionFrame;
  }
  resizeBoxOver(event) {
    super.resizeBoxOver(event);
    this.transcriptionFrame.style.width = this.boxWidth+"px";
  }

  addSpan(x) {
    if (codeFormat == "html") {
      var span = document.createElement("span");
      span.className = x;
    } else {
      var span = document.createElement(x);
    }
    return span;
  }
  addShortcut(x) {
    var range = window.getSelection().getRangeAt(0);
    var b = document.createElement('span');
    if (document.getElementById('codeBtn').innerHTML.match('xml')) {
      x = x.replace(/<\/.*?>/g,"</span>");
      x = x.replace(/<([^\/]*?)>/g,"<span class=\"$1\">");
    } else {
      x = x.replace(/</g,"&lt;");
      x = x.replace(/>/g,"&gt;");
    }
    b.innerHTML = x;
    b.innerHTML += "\u200B";
    if (!window.getSelection().isCollapsed) {
      range.extractContents();
    }
    range.insertNode(b);
    range.collapse(false); //moves cursor to end of range
    window.getSelection().removeAllRanges();
    window.getSelection().addRange(range);

    this.transcriptionFrame.focus();
    return true;
  }
  addTag(x) {
    var range = window.getSelection().getRangeAt(0);
    if (codeFormat == "html") {
      if (this.unselectRange(range,x) === false) {
        var node = this.addSpan(x);
        node.appendChild(range.extractContents());
        range.insertNode(node);
        range.setStartAfter(node);
        var emptySpaceAfterTag = document.createTextNode("\u200B");
        range.insertNode(emptySpaceAfterTag);
      }
      this.cleanTags(x);
    } else if (codeFormat == "xml") {
        var node = document.createTextNode("<"+x+">");
        range.insertNode(node);
        range.collapse(false);
        var node = document.createTextNode("</"+x+">");
        range.insertNode(node);
    }
    range.collapse(false);
    this.transcriptionFrame.focus();
    detectChange();
    return true;
  }
  unselectRangeNew(range,x) {
    var unselect = false;
    var range = window.getSelection().getRangeAt(0);
    var rangeNode = this.addSpan("rangeTest");
    rangeNode.appendChild(range.extractContents());
    range.insertNode(rangeNode);

    var nodeAry = this.transcriptionFrame.getElementsByClassName(x);
    for (var n = 0; n < nodeAry.length; n++) {
      var node = nodeAry[n];
    }


    rangeNode.parentNode.removeChild(rangeNode);
    return unselect;
  }
  unselectRangeTest(range,x) {
    this.transcriptionFrame.normalize(); 
    var orig = this.transcriptionFrame.innerHTML;
    var node;
    while (node = this.transcriptionFrame.querySelector("."+x+" ."+x)) {
      node.outerHTML = node.innerHTML;
    } 
  }
  unselectRange(range,x) {
    var unselect = false;
    var nodeAry = this.transcriptionFrame.getElementsByClassName(x);
    for (var n = 0; n < nodeAry.length; n++) {
      var node = nodeAry[n];
      var rangeNode = document.createRange();
      rangeNode.selectNodeContents(node);

      //move range start inside node, where appropriate
      if (range.compareBoundaryPoints(Range.START_TO_START,rangeNode) == -1) {
        var rangeTest = range.cloneRange();
        rangeTest.setStart(node,0);
        if (range.toString() === rangeTest.toString()) {
          range.setStart(rangeNode.startContainer,rangeNode.startOffset);
        }
      }
      //move range end inside node, where appropriate
      if (range.compareBoundaryPoints(Range.END_TO_END,rangeNode) == 1) {
        var rangeTest = range.cloneRange();
        rangeTest.setEnd(rangeNode.endContainer, rangeNode.endOffset);
        if (range.toString() === rangeTest.toString()) {
          range.setEnd(rangeNode.endContainer, rangeNode.endOffset);
        }
      }
      
      var rangeParent = range.commonAncestorContainer;
      if (node.contains(rangeParent)) {
        //range is inside existing node, so unselect
        //split the node into two ranges, from beginning of node to beginning of range, and from end of range to end of node
        //cleanTags method will remove empty/redundant tags

        var range1 = document.createRange();
        range1.setStart(rangeNode.startContainer,rangeNode.startOffset);
        range1.setEnd(range.startContainer,range.startOffset);

        var range2 = document.createRange();
        range2.setStart(range.endContainer,range.endOffset);
        range2.setEnd(rangeNode.endContainer,rangeNode.endOffset);

        var node1 = this.addSpan(x);
        node1.appendChild(range1.extractContents());
        if (node1.textContent != "") { range1.insertNode(node1); }

        var node2 = this.addSpan(x);
        node2.appendChild(range2.extractContents());
        if (node2.textContent != "") { range2.insertNode(node2); }

        node.outerHTML = node.innerHTML;
        unselect = true;
      }
      rangeNode.detach();
    }
    return unselect;
  }
  cleanTags(x) {
    //this.transcriptionFrame.innerHTML = this.transcriptionFrame.innerHTML.replace(/\u200B/g,"");
    //zero-width spaces were throwing things off
    //they are needed to type a word, tag it, and then continue typing outside the tag
    //check for them within the functions below instead of a global search and replace

    //because of a cascade effect, changed this function to check all tags, not just the tag applied

    this.transcriptionFrame.normalize(); 
    var orig = this.transcriptionFrame.innerHTML;
    var node;
    
    //test for adjacent siblings
    //var nodeAry = this.transcriptionFrame.getElementsByClassName(x);
    var nodeAry = this.transcriptionFrame.getElementsByTagName("span");
    for (var n = 0; n < nodeAry.length; n++) {
      var node = nodeAry[n];
      if (node.className) {
        var className = node.className;
        var prev = node.previousSibling;
        while (prev && prev.nodeType == 3 && prev.textContent == "\u200B") {
          prev = prev.previousSibling;
        }
        if (prev && prev.nodeType == 1 && prev.className.indexOf(className) > -1) {
          prev.appendChild(node);
        }
      }
    } 
    
    //test for redundant children
    //while (node = this.transcriptionFrame.querySelector("."+x+" ."+x)) {
    var nodeAry = this.transcriptionFrame.getElementsByTagName("span");
    for (var n = 0; n < nodeAry.length; n++) {
      var node = nodeAry[n];
      if (node.className) {
        var childAry = node.getElementsByClassName(node.className);
        for (var c = 0; c < childAry.length; c++) {
          childAry[c].outerHTML = childAry[c].innerHTML;
        } 
      } 
    } 
    this.transcriptionFrame.normalize(); 

    //test for empty nodes
    //var nodeAry = this.transcriptionFrame.getElementsByClassName(x);
    var nodeAry = this.transcriptionFrame.getElementsByTagName("span");
    for (var n = 0; n < nodeAry.length; n++) {
      var node = nodeAry[n];
      var text = node.textContent;
      text = text.replace("\u200B","");
      if (text == "") {
        node.parentNode.removeChild(node);
      }
    }
    this.transcriptionFrame.normalize(); 

    //combine identical siblings
    var nodeAry = this.transcriptionFrame.getElementsByTagName("span");
    for (var n = 0; n < nodeAry.length; n++) {
      var node = nodeAry[n];
      var prev = node.previousSibling;
      if (prev && prev.nodeType == 1) {
        if (node.className == prev.className) {
          prev.appendChild(node);
        }
      }
    } 

    this.transcriptionFrame.normalize(); 
    if (orig != this.transcriptionFrame.innerHTML) { this.cleanTags(x); }
    //this.transcriptionFrame.innerHTML = this.transcriptionFrame.innerHTML.replace("</span>","</span>\u200B");

  }
  cleanEmptyTags() {
    this.transcriptionFrame.normalize(); 
    var orig = this.transcriptionFrame.innerHTML;
    var nodeAry = this.transcriptionFrame.getElementsByTagName("*");
    for (var n = 0; n < nodeAry.length; n++) {
      var node = nodeAry[n];
      if (node.nodeName == "BR") {
      } else if (node.textContent == "") {
        node.parentNode.removeChild(node);
      }
    } 
    this.transcriptionFrame.normalize(); 
    if (orig != this.transcriptionFrame.innerHTML) { this.cleanEmptyTags(); }
  }
  cleanSelfClosingTags(text) {
    //Self-closing tags are ok in XML, but not HTML. InnerHTML will add additional closing tags.
    //Ignore <lb/> and <br/>, since those will be handled separately.
    return text.replace(/<(?!lb|br)([^ >]+)([^>]*?)\/>/ig,"<$1$2></$1>");
  }
}
